# Limine-Snapper-Sync

### Contents

- [Description](#description)
- [Screenshots](#screenshots)
- [Requirements](#requirements)
- [Installation](#installation)
- [Boot Configuration](#boot-configuration)
- [Tool Configuration](#tool-configuration)
- [Implementation](#implementation)
- [Donate](#donate)
- [Planned Version 2.0.0](#donate)

## Description

The tool syncs [Limine](https://github.com/limine-bootloader/limine) boot entries with Snapper snapshots (Btrfs snapshots).

**What it does:**

* **Adding Snapshots:**
When a Snapper snapshot is created, a snapshot boot entry with its matching kernel versions is added to Limine.

* **Deleting Snapshots:**
Deleting a Snapper snapshot also removes the related snapshot boot entry and kernel files.

* **Restoring Snapshot:**
The tool supports `rsync` or `btrfs` to restore a selected snapshot with matching kernel versions.

* **Backup Entry:**
After restoring, a "backup" entry is created and added to Limine, allowing you to revert or remove it later.

* **Boot Files Repair:**
Automatically detects and repairs a corrupted bootable snapshot file by replacing it with a matching healthy file
during new snapshot creation.
Additionally, it logs hardware issues if two hashes of the same bootable file don’t match.

This tool supports custom Snapper layouts but not arbitrary Btrfs layouts, which is why it’s named
"limine-**snapper**-sync" instead of "limine-**btrfs**-sync".

**Note:** This tool does not create boot entries when you install a new kernel.
If you need that functionality, use [limine-entry-tool](https://gitlab.com/Zesko/limine-entry-tool),
which helps create or remove boot entries, just like `grub-install` and `grub-mkconfig` for Grub.

#### Why Limine-Snapper-Sync?

No other solution supports all the following:

* **Faster encrypted snapshot boot:**
    - Bootable files (kernel and initramfs/initrd, or UKI) are stored outside a fully encrypted partition,
      allowing faster decryption and quicker boot times.
    - Supports Plymouth graphical password prompts during boot for a more polished user experience.
* **Multiple boot support:**
    - Each mutable Linux distro can have its own snapshot entries in separate boot submenus,
      making it easier to select OS or snapshot entry.
* **Efficient design:**
    - Avoids reliance on third-party drivers (e.g., filesystems, LUKS2, different encryption algorithms, LVM, etc.);
      all of this is handled by kernel, initramfs, or UKI, improving compatibility and reliability.
* **Instant snapshot updates:**
    - Automatically updates the snapshot list in real time without scanning all subvolumes or snapshots.
* **Rollback and "Backup" snapshot support:**
    - Allows quick system restore and adds a "backup" snapshot boot entry.
* **Automatic checksum correction & faulty RAM detection:**
    - Detects faulty RAM or CPU cache by comparing two independently generated hashes.
    - Corrects checksum errors caused by faulty RAM.

Therefore, this tool was created to provide a faster, more efficient, and easier-to-manage snapshot solution,
thanks to Limine's design.

#### Limitations

* **Old snapshots before installing the tool**:
    - Snapshots created before installing this tool won’t be added to the Limine bootloader because they’re missing
      their kernel files.
* **Limited Btrfs layout support:**
    - The tool works only with default or custom Snapper layouts.

#### There are five available commands, each without any options:

* `limine-snapper-sync` updates Limine snapshot entries with current Snapper snapshots (similar to `grub-btrfs` for GRUB).
* `limine-snapper-list` displays all available snapshots in Limine's boot menu.
* `limine-snapper-info` shows details about versions, total bootable snapshots, kernel verification status.
* `limine-snapper-restore` restores a selected snapshot to your system.
* `limine-snapper-watcher` should be run **once** at startup to auto-sync Limine with Snapper snapshots.

## Screenshots:

Choose between two different Limine menu layouts:

### Snapshots-menu inside OS entry

The "Snapshots" menu is integrated within the OS entry.
![image](screenshots/1.jpg)

Expand it to view snapshot details such as **ID**, **date** and **description**.
Each Linux distro can have its own separate snapshot list.

![image](screenshots/2.jpg)

### Snapshots-menu outside the OS entry

The "Snapshots" menu is listed separately from the OS entry.
![image](screenshots/7.jpg)

After booting from a snapshot, a notification appears, allowing one-click restoration.

![image](screenshots/3.jpg)

Clicking "Restore now" provides more details about the restore process.

![image](screenshots/4.jpg)

If a display manager (e.g., GDM) fails on a read-only snapshot, switch to TTY and run: `limine-snapper-restore`

Run `limine-snapper-info` to verify bootable snapshot files.

![image](screenshots/6.jpg)

## Requirements

* A FAT32 boot partition of at least **2 GiB** is recommended, depending on how many snapshots and kernel versions you keep.
* Limine version **8 or newer**
* `snapper`
* `btrfs-progs`
* `inotify-tools` for monitoring when snapshots are created or deleted.
* Standard coreutils: `cp`, `mv`, `rm`, `sha256sum`
* Standard util-linux: `logger`, `mount`, `umount` and `flock`
* Java version 17 or newer
`maven` for building the project

## Installation:

### Arch Linux:

**🔄 Switching from GRUB, systemd-boot, or rEFInd to Limine?**<br>
**Tip:**
Limine setup allows you to keep your existing EFI bootloader as a fallback.<br>
It won’t overwrite your existing EFI bootloader unless you explicitly choose to replace it.<br>
Just follow the **easy setup with `limine-entry-tool`** below if you are unsure.

#### 🚀 Easy setup with limine-entry-tool

1. Choose one of the two initramfs integration packages for [limine-entry-tool](https://gitlab.com/Zesko/limine-entry-tool):
   * For **Dracut** users: Install [limine-dracut-support](https://aur.archlinux.org/packages/limine-dracut-support/)
   * For **mkinitcpio** users: Install [limine-mkinitcpio-hook](https://aur.archlinux.org/packages/limine-mkinitcpio-hook/)
2. Install [limine-snapper-sync](https://aur.archlinux.org/packages/limine-snapper-sync/) from AUR.
3. Check if an ESP path is detected by running: `bootctl --print-esp-path`:
    * If detected, no further configuration is needed.
    * If not detected, create or edit `/etc/default/limine` and set `ESP_PATH=` to any mounted larger FAT32 boot partition of your choice.
      Then run `limine-install` and `limine-update` to verify everything is working.
4. Enable automatic syncing:
```bash
systemctl enable --now limine-snapper-sync.service
```
5. (Optional) Install `snap-pac` to trigger Snapper to create snapshots during system updates.
6. (Optional) Install [journalctl-desktop-notification](https://gitlab.com/Zesko/journalctl-desktop-notification) from AUR for desktop notifications about full ESP usage or
detected hardware issues.

**Tip:** After changing your config in `/etc/default/limine`, just run `limine-update` to apply the changes.
It will rebuild your initramfs or UKI, similar to reinstalling the kernel.

**or**

#### Manual setup without limine-entry-tool

1. Make sure Limine can boot your system by following the [Arch Wiki:Limine](https://wiki.archlinux.org/title/Limine#Limine-Snapper-Sync_setup).
1. Properly configure `limine.conf` to work with this tool. See [Boot Configuration](#Boot-Configuration) for details.
1. Install [limine-snapper-sync](https://aur.archlinux.org/packages/limine-snapper-sync/) from AUR.
1. Check if an ESP path is detected: `bootctl --print-esp-path`
    * If not detected, create and edit `/etc/default/limine` and set `ESP_PATH=` to your FAT32 boot partition.
1. Create a root snapshot using `snapper`, then test `limine-snapper-sync` to see if
   `Saved successfully: $ESP_PATH/limine.conf`.
1. If everything works, enable automatic syncing:
```bash
systemctl enable --now limine-snapper-sync.service
```

**Note:** If it shows an error that command `limine-reset-enroll` and `limine-enroll-config` not found,
these commands are provided by some `limine-entry-tool` package, which is not installed.
This error can be safely ignored if you do not use them. Alternatively, to prevent this error, 
edit `/etc/limine-snapper-sync.conf` to remove:
```
COMMANDS_BEFORE_SAVE="limine-reset-enroll"
COMMANDS_AFTER_SAVE="limine-enroll-config"
```
------

### For other Linux distributions:

This tool should work on any Linux distribution, with or without systemd.
However, there are no pre-built packages for various package managers.
Package maintainers are welcome to create them.

**Note:** This setup below is not suitable for NixOS and some immutable distros as they do not
follow [FHS](https://en.wikipedia.org/wiki/Filesystem_Hierarchy_Standard).

Manual installation is required:

#### **For systemd distros:**

* Ensure that Limine can boot your system.
* Clone the repository:

```shell
git clone https://gitlab.com/Zesko/limine-snapper-sync.git
cd limine-snapper-sync
```

* Build the project

```shell
mvn clean package
```

* Install the necessary files:

```shell
cp target/limine-snapper-sync.jar /usr/share/java/
cp install/arch-linux/usr/bin/* /usr/local/bin/
cp -r install/arch-linux/usr/lib/limine /usr/lib/
cp -r install/arch-linux/etc/* /etc/
cp -r install/arch-linux/usr/lib/systemd/* /etc/systemd/
```

* Install `libnotify` or [dunst](https://github.com/dunst-project/dunst) for desktop notifications on X11 or Wayland.
* Properly configure `limine.conf` to work with this tool. See [Boot Configuration](#Boot-Configuration) for details.
* Test to run `limine-snapper-sync`. If successful, you should see: `Saved successfully: $ESP_PATH/limine.conf`.
* Enable the service to start syncing automatically:

```shell
systemctl enable --now limine-snapper-sync.service
```

#### **For non-systemd distros:**

* Ensure that Limine can boot your system.
* Clone the repository:

```shell
git clone https://gitlab.com/Zesko/limine-snapper-sync.git
cd limine-snapper-sync
```

* Build the project

```shell
mvn clean package
```

* Install the necessary files:

```shell
cp target/limine-snapper-sync.jar /usr/share/java/
cp install/arch-linux/usr/bin/* /usr/local/bin/
cp -r install/arch-linux/usr/lib/limine /usr/lib/
cp install/arch-linux/etc/xdg/autostart/*.desktop /etc/xdg/autostart/` # Adjust based on ${XDG_CONFIG_DIRS}
cp install/arch-linux/etc/*.conf /etc/
````

* Install `libnotify` or [dunst](https://github.com/dunst-project/dunst) for desktop notifications on X11 or Wayland.
* Properly configure `limine.conf` to work with this tool. See [Boot Configuration](#Boot-Configuration) for details.
* Test to run `limine-snapper-sync`. If successful, you should see: `Saved successfully: $ESP_PATH/limine.conf`.
* Create your own service or script to run `/usr/local/bin/limine-snapper-watcher` on every reboot.

## Boot Configuration:

**Tip:** Boot configuration is automated by some `limine-entry-tool`-based package.
This allows you to skip the manual configuration below unless custom options are needed.

### Manual Limine configuration: 
**Edit `$ESP_PATH/limine.conf`:**

1. **Add a kernel entry:**

   You **must** add at least one kernel entry in the format `//<Kernel name>` (e.g., `//Linux LTS`)
   within an OS entry block `/<OS name>`.

2. **Choose one snapshot layout:** (Optional)

    * Default: Add `//Snapshots` inside the OS entry block.

   **or**

    * Alternative: Add `/Snapshots` outside the OS entry block.

The keyword (`//Snapshots` or `/Snapshots`) tells the tool where snapshot entries should be placed.

#### Example 1:

If bootable kernel files are stored in `/efi/`, `/boot/`, `/boot/efi/`, or any other path used as your boot partition:

```
/Arch Linux

  //Kernel LTS
  protocol: linux
  module_path: boot():/initramfs-linux-lts.img
  kernel_path: boot():/vmlinuz-linux-lts
  kernel_cmdline: rw rootflags=subvol=/@ root=UUID=068589c1-d9be-407c-ad86-101aa0912825
  
  //Kernel
  protocol: linux
  module_path: boot():/initramfs-linux.img
  kernel_path: boot():/vmlinuz-linux
  kernel_cmdline: rw rootflags=subvol=/@ root=UUID=068589c1-d9be-407c-ad86-101aa0912825

  //Snapshots
```

#### Example 2:

If bootable kernel files are stored in `/efi/d647abee5e264f469d7a38b751d9d1db/kernel-version/`
like where systemd-boot (`kernel-install`) stores them:

```
/+Arch Linux
comment: Any comment
comment: machine-id=d647abee5e264f469d7a38b751d9d1db

  //Mainline kernel
  protocol: linux
  module_path: boot():/d647abee5e264f469d7a38b751d9d1db/kernel-version/initrd
  kernel_path: boot():/d647abee5e264f469d7a38b751d9d1db/kernel-version/linux
  kernel_cmdline: rw rootflags=subvol=/@ root=UUID=068589c1-d9be-407c-ad86-101aa0912825

  //Mainline kernel fallback
  protocol: linux
  module_path: boot():/d647abee5e264f469d7a38b751d9d1db/kernel-version/initrd-fallback
  kernel_path: boot():/d647abee5e264f469d7a38b751d9d1db/kernel-version/linux
  kernel_cmdline: rw rootflags=subvol=/@ root=UUID=068589c1-d9be-407c-ad86-101aa0912825

/Snapshots
```
**Note:** Bootable files must be placed in the boot partition alongside `limine.conf`.
Version 1.y.x only supports the `boot():` function.
The upcoming version 2.0.0 will introduce support for `guid(<GPT partition UUID>):`,
allows bootable files to be stored in a different partition.

______

## Tool Configuration

The config file is located at `/etc/default/limine` (**Recommend**) or alternatively at `/etc/limine-snapper-sync.conf`

**Note:**

- `/etc/default/limine` takes precedence over `/etc/limine-snapper-sync.conf`.
- When booting into a read-only snapshot where `/etc/*` cannot be modified,
`/etc/default/limine` is automatically copied to the temporary location `/tmp/limine-snapper-sync.conf`,
allowing temporary modifications.

Copy this to `/etc/default/limine` and adjust the settings as needed.

```properties
### OS Entry Targeting:
### The tool does not automatically know the custom OS name used in /<ESP>/limine.conf, which can contain multiple OS names for multi-boot scenarios.
### You need to specify the corresponding OS name here to match the name used in /<ESP>/limine.conf.
### Alternatively, you can add "comment: machine-id=<machine-id>" to your OS entry block in /<ESP>/limine.conf. 
### The machine-ID helps the tool to automatically target the correct OS entry. Changing the OS name does not matter.
TARGET_OS_NAME="Arch Linux"


### Boot Partition Path:
### Specify the mount path of your boot partition if you are running on legacy BIOS, ARM, or RISC-V.
### Note:
### If you are using systemd and UEFI, the path is detected automatically via 'bootctl --print-esp-path'.
### However, specifying the ESP path manually can improve performance and reliability.
#ESP_PATH="/boot"


### Max Snapshot Entries:
### Set the maximum number of snapshot entries for your OS. This depends on the boot partition size and how many kernels will be installed.
### Note: The tool uses a deduplication function to prevent copying kernel versions that already exist.
MAX_SNAPSHOT_ENTRIES=8


### Boot Partition Usage Limit:
### Set a usage limit for your boot partition as a percentage between 1 and 99. The default value is 85.
### The tool will stop creating new snapshot entries when this limit is reached.
### Tip for Arch Linux users: Install "journalctl-desktop-notification" from AUR, it will notify you when the usage limit is exceeded.
LIMIT_USAGE_PERCENT=85


### Make Snapshots Writable (Not Recommended)
### When set to 'yes', created snapshots will be writable. (yes|no)
### This may be required by some programs or display managers like GDM that need write access at boot when no overlayfs is used.
###
### WARNING:
### Writable snapshots can be modified and will no longer reflect the original 1:1 state.
### Enable only if absolutely necessary.
### IMPORTANT:
### If you boot into a writable snapshot, you should click "Restore now" immediately.
#SNAPSHOT_WRITABLE=no


### Snapper Configuration Name:
### Specify the Snapper configuration name. If not set, the tool will automatically detect it from "snapper list-configs".
### Specifying it manually improves performance by avoiding automatic detection overhead.
SNAPPER_CONFIG_NAME="root"


### Root Subvolume Path:
### Set the path to your root subvolume. The default value for some distros is "/@".
ROOT_SUBVOLUME_PATH="/@"


### Root Snapshot Path:
### Specify the path to your root snapshots. The default is "/@/.snapshots", which is used for the default Snapper layout.
### Note: The tool supports Snapper with any custom layout within the same filesystem, but not random Btrfs layouts without Snapper.
ROOT_SNAPSHOTS_PATH="/@/.snapshots"


### Restore Method Selection:
### Choose whether to enable rsync for restoration. If enabled, the tool will prompt you to choose between rsync or btrfs during the restore process.
### Advantage of rsync: The subvolume ID remains unchanged without "messing up" the Btrfs metadata.
### Disadvantage of rsync: Slower compared to Btrfs, which directly creates a new subvolume from the selected snapshot.
### "yes": Enables the prompt for choosing rsync or btrfs.
### "no":  Disables the prompt, defaulting to btrfs for "one-click-restore."
ENABLE_RSYNC_ASK=no


### Btrfs UUID:
### Specify the UUID of your Btrfs filesystem, used only during the restore process.
### If not set, the tool automatically determines the UUID during initialization and saves it in 'snapshots.json'. 
### If the saved UUID becomes outdated, set it manually.
#UUID=


### Authentication Method for Restore:
### Specify an authentication method for restore process. Options include 'sudo', 'doas', 'pkexec', `run0` or another method of your choice.
### If not set, 'pkexec' is used for Wayland/X11 environments, and 'sudo' for TTY.
#AUTH_METHOD=sudo


### Specify a terminal app and its argument to open the console UI for running the restore process with details.
### For example, "konsole -e" for KDE or "gnome-terminal -- bash -c" for GNOME.
#TERMINAL=konsole
#TERMINAL_ARG="-e"


### Snapshot Entry Formatting:
### Set the number of spaces to indent each line in all snapshot entries within /<ESP>/limine.conf.
SPACE_NUMBER=5


### Snapshot Name Format:
### Choose the format for how snapshot entries look in the Limine bootloader:
### 0. ID=111 2023-12-20 10:59:59 (default)
### 1. 111│2023-12-20 10:59:59
### 2. 111 │ 2023-12-20 10:59:59
### 3. 2023-12-20 10:59:59│111
### 4. 2023-12-20 10:59:59 │ 111
### 5. 2023-12-20 10:59:59
### 6. 111
SNAPSHOT_FORMAT_CHOICE=2


### Hash Function for Deduplication and Checksum:
### Select a hash function to avoid copying duplicate boot files. Options are "blake3", "sha1", or "sha256".
### WARNING: Changing the hash function will not affect files named with the previous hash function, potentially increasing duplication.
### Note: "blake3" must be installed manually if chosen.
HASH_FUNCTION=sha256


### Notification Icon:
### Set a custom icon to display in notifications when booting into a read-only snapshot and prompting for restoration.
NOTIFICATION_ICON="/usr/share/icons/hicolor/128x128/apps/LimineSnapperSync.png"


### Automatic Config Backup:
### Create a backup of 'limine.conf' and 'snapshots.json' before saving changes if they are older than the specified time (in hours).
### If not set, the default threshold is 8 hours.
BACKUP_THRESHOLD=8


### Custom Java Binary Path
### Set this to specify a custom Java binary path instead of using the system default.
### If left unset, the default system Java will be used.
#JAVA_BIN_PATH="/usr/lib/jvm/java-21-openjdk/bin"


### Quiet Mode
### Set to "yes" to hide all success messages and avoid log spam. (yes|no)
#QUIET_MODE=no


### Commands Before Save
### Run custom commands before saving 'limine.conf'.
### If using multiple commands, separate them with '&&' or '||'.
### Note: Use single quotes (') for command arguments when required, not double quotes (") as escape characters.
### Example: 'limine-reset-enroll' command is provided by some limine-entry-tool package.
COMMANDS_BEFORE_SAVE="limine-reset-enroll"


### Commands After Save
### Run custom commands after saving 'limine.conf'.
### If using multiple commands, separate them with '&&' or '||'.
### Note: Use single quotes (') for command arguments when required, not double quotes (") as escape characters.
### Example: 'limine-enroll-config' command is provided by some limine-entry-tool package.
COMMANDS_AFTER_SAVE="limine-enroll-config"

```

## Implementation:

#### First Start

* **Machine ID Generation:** If `/etc/machine-id` is missing, the tool generates a new ID for unique OS identification, which is useful for multi-boot setups.
* **Configuration Verification:** The tool checks your Btrfs layout and config. If they are invalid, the tool will not proceed.
* **Metadata Recovery:** If a `snapshots.json` file becomes corrupted, the tool automatically restores a backup `snapshots.json.old`.
* **Limine Entry Parsing:** All Limine boot entries are read and converted into a tree structure to simplify management.
* **Atomic Locking:**  A mutex ensures that only one instance of the tool runs at a time to avoid conflicts.

#### Snapshot Creation and Deletion

* **Real-Time Monitoring:** Uses `inotifywait` to monitor the `/.snapshots/` directory for snapshot creation and deletion events.
* **Snapshot Integration:** When a new Snapper snapshot is created, the tool copies the current OS boot entry and associated
  kernel files into a new Limine snapshot entry.
* **Kernel Deduplication**: Avoids redundant copying by re-using kernel files already present on the boot partition.
* **Metadata Update:** Updates `snapshots.json` with snapshot-to-kernel associations.
* **Snapshot List Update:** Automatically updates the list of snapshot entries in `limine.conf`.
* **Snapshot Removal:** Removes snapshot-related boot entries and deletes kernel files if no other snapshot depends on them.
* **File Integrity Repair:** Replaces corrupted boot files on the boot partition with a verified good during snapshot creation.

#### Restore Process

* **Pre-Restore Validation:** Btrfs layout and configuration are verified again before restoring. If invalid, the restore is aborted.
* **Subvolume Restore:** A new subvolume is created from a selected snapshot. All child subvolumes are moved to it.
* **Rsync Restore:** `rsync` restores files from a selected snapshot, excluding bind mounts and child subvolumes. Files not present in the snapshot are deleted.
* **Kernel Backup:** Current "broken" kernel files are backed up, and a "backup" entry is added to the Snapper list.
* **Kernel Restore:** Kernel files are restored from the `limine_history` directory to their correct location.
* **Boot Entry Restore:** The original boot entry from `snapshots.json` is written back to the OS section in `limine.conf`.

______________________

## Donate

If you would like to support my work on the planned version 2.0.0,
a complete rewrite in Kotlin with some advantages and improvements over version 1, feel free to donate:

[<img src="https://gitlab.com/Zesko/resources/-/raw/main/kofiButton.png" width=110px>](https://ko-fi.com/zeskotron)
&nbsp;[<img src="https://gitlab.com/Zesko/resources/-/raw/main/PayPalButton.png" width=160px/>](https://www.paypal.com/donate/?hosted_button_id=XKP36D62AY5HY)


#### Planned version 2.0.0

- **Support for smaller ESP sizes** by moving bootable snapshot files to a separate FAT32 partition.
- **A new option`NO_SNAPSHOT_KERNELS=`:** allows excluding specific kernels (e.g., `*fallback`) from new snapshots.
- **Btrfs layout changes for Snapper:** All old kernel parameters stored in metadata will be updated automatically.
- **Pinned snapshots:** They always stay and are not affected by the `MAX_SNAPSHOT_ENTRIES` cleanup setting.
- **Bcachefs support** will be added once Snapper is ready — see [TODOs](https://github.com/openSUSE/snapper/blob/master/doc/bcachefs.txt).
- **Enhanced performance** by running multiple processes in parallel using optimized Kotlin code and standard libraries.
- **Smaller binary size** than version 1.
- **Full compatibility** with version 1.